/*!
 *@file		parcel.cc
 *@author	Craig Ruff
 *@date		July 31, 2002
 *@version	2
 *
 *@brief  Parcel implementation.
 *
 * Copyright 2002 University Corporation for Atmospheric Research
 * All Rights Reserved
 */

#include <pthread.h>
#include "package.h"
#include "packet.h"
#include "parcelp.h"
#include "parmfile.h"
#include "quoting.h"
#include "svec.h"
#include "svecpack.h"

using std::string;

namespace Libtools {

Parcel::Parcel()
{
	rep = new Rep();
}

Parcel::Parcel(const Package& package)
{
	rep = new Rep();

	//
	// BundlePackage will cause memory allocation.
	// Catch an out of memory exception so that we
	// don't leave this object in a partially
	// constructed state.
	//
	try {
		Merge(package);

	} catch (std::bad_alloc) {
		delete rep;
		throw; // re-throw
	}
}

Parcel::Parcel(const Parcel& p)
{
	rep = p.rep->Use();
}

Parcel::Parcel(const Packet& p)
{
	rep = new Rep();
	Merge(p);
}

Parcel::Parcel(IParmfile& ip)
{
	rep = new Rep();
	Merge(ip);
}

Parcel&
Parcel::Clear(void)
{
	rep = rep->Clear();
	return(*this);
}

Parcel&
Parcel::Assign(const Parcel& p)
{
	if (rep == p.rep)
		return(*this);		// protect against x = x

	Rep*	t = p.rep->Use();
	if (rep->Release() == 0)
		delete rep;

	rep = t;
	return(*this);
}

bool
Parcel::Get(const std::string& key, std::string& s, const std::string& d) const
{
	const std::string*	v = Get(key);
	if (v == 0) {
		s = d;
		return(false);
	}

	s = *v;
	return(true);
}

bool
Parcel::Get(const std::string& key, Parcel& p) const
{
	const std::string*	v = Get(key);
	if (v == 0)
		return(false);

	p.Assign(Package(*v));
	return(true);
}

Parcel&
Parcel::Merge(const Parcel& p)
{
	if (rep == p.rep)
		return(*this);

	Rep::const_iterator i = p.rep->begin();
	Rep::const_iterator e = p.rep->end();

	while (i != e) {
		Set(i->first, i->second);
		++i;
	}

	return(*this);
}

Parcel::~Parcel()
{
	if (rep->Release() == 0)
		delete rep;
}

Parcel&
Parcel::Set(const std::string& key, const std::string& value)
{
	rep = rep->Set(key, value);
	return(*this);
}

Parcel&
Parcel::Erase(const std::string& key)
{
	rep = rep->Erase(key);
	return(*this);
}

const std::string*
Parcel::Get(const std::string& key) const
{
	Rep::const_iterator p = rep->find(key);

	if (p != rep->end())
		return(&p->second);

	return(0);
}

bool
Parcel::Get(const std::string& key, Svec& s) const
{
	const std::string*	v = Get(key);
	if (v == 0)
		return(false);

	s.Assign(SvecPack(*v));
	return(true);
}

Parcel&
Parcel::Merge(const Package& p)
{
	Package::const_iterator	b = p.begin();
	Package::const_iterator	e = p.end();

	while (b != e) {
		rep->Set(b->first, b->second);
		++b;
	}

	return(*this);
}

Parcel&
Parcel::Merge(const Packet& p)
{
	Packet::const_iterator	b = p.begin();
	Packet::const_iterator	e = p.end();
	const char*		key;
	const char*		value;

	while (b != e) {
		key = *b;
		if (++b == e) {
			value = "";

		} else {
			value = *b;
			++b;
		}

		rep->Set(key, value);
	}

	return(*this);
}

Parcel&
Parcel::Merge(IParmfile& ip)
{
	IParmfile::iterator	b = ip.begin();
	IParmfile::iterator	e = ip.end();

	while (b != e) {
		if (!b->first.empty())
			rep->Set(b->first, b->second);

		++b;
	}

	return(*this);
}

const Parcel&
Parcel::Output(std::ostream& s, char sep) const
{
	for (Rep::const_iterator i = rep->begin(); i != rep->end(); ++i) {
		s << Quote(i->first) << '=' << Quote(i->second) << sep;
	}

	return(*this);
}

bool
Parcel::Foreach(Visitor& v) const
{
	for (Rep::const_iterator i = rep->begin(); i != rep->end(); ++i) {
		if (v(i->first, i->second))
			return(true);
	}

	return(false);
}

Parcel&	
Parcel::Set(const std::string& key, const Parcel& subp)
{
	Set(key, Package(subp));
	return(*this);
}

Parcel&	
Parcel::Set(const std::string& key, const Svec& s)
{
	Set(key, SvecPack(s));
	return(*this);
}

} // end namespace Libtools
