/*!
 *@file		parcel.h
 *@author	Craig Ruff
 *@date		March 22, 2005
 *@version	3
 *
 *@brief Parcel public interface definitions.
 *
 * Copyright 2005 University Corporation for Atmospheric Research
 * All Rights Reserved
 */

#ifndef LIBTOOLS_PARCEL_H
#define LIBTOOLS_PARCEL_H

#include <string>
#include <ostream>
#include <sstream>

namespace Libtools {

class Package;
class Packet;
class IParmfile;
class Svec;

/*!
 *@brief Associative key/value memory.
 *
 * A Parcel stores key/value pairs in an associative memory.  The key and
 * values types are 8-bit character strings that do not contain embedded NUL
 * characters (binary zeros).
 *
 * The Parcel API is similar to that provided by the older C libtools library
 * PARCEL datatype.  A Parcel provides automatic memory management and
 * an improved convenience as compared to the equivalent C libtools
 * PARCEL type.
 *
 * A Package or a Parmfile are the human readable forms of a Parcel.
 */
class Parcel
{
private:
	/*!
	 *@internal
	 *
	 * Rep is the internal representation that is
	 * not visible to clients of the interface.
	 */
	class Rep;
	Rep	*rep;

	/*!
	 *@internal
	 *@brief Helper class to allow subscripting along with operator[].
	 */
	class ItemRef
	{
	private:
		Parcel&			parcel;
		const std::string&	key;

		ItemRef();
		ItemRef& operator=(const ItemRef&);

	public:
		/*!
		 *@brief Constructor
		 *
		 *@param[in] p	The parent Parcel.
		 *@param[in] k	The key to use.
		 */
		ItemRef(Parcel& p, const std::string& k)
			: parcel(p), key(k)
		{}

		/*!
		 *@brief Read an item.
		 *
		 *@retval 0	@p key was not found.
		 *@retval !0	The pointer to the value string.
		 */
		operator const std::string*() const
		{
			return(parcel.Get(key));
		}

		/*!
		 *@brief Assign an item.
		 *
		 *@param[in] value	The value to assign to @p key.
		 */
		template<typename T>
		void
		operator=(const T& value)
		{
			parcel.Set(key, value);
		}
	};


public:
	//!@name Constructors
	//@{
	/*!
	 *@brief Construct an empty Parcel.
	 */
	Parcel();

	/*!
	 *@brief Construct a copy of another Parcel.
	 *
	 *@param[in] p	The Parcel to copy.
	 */
	Parcel(const Parcel& p);

	/*!
	 *@brief Construct from a Package.
	 *
	 *@param[in] p	The Package.
	 */
	explicit
	Parcel(const Package& p);

	/*!
	 *@brief Construct from a Packet.
	 *
	 *@param[in] p	The Packet.
	 */
	explicit
	Parcel(const Packet& p);

	/*!
	 *@brief Construct from an IParmfile.
	 *
	 *@param[in] ip	The IParmfile.
	 */
	explicit
	Parcel(IParmfile& ip);
	//@}

	//!@name Accessor Methods
	//@{
	/*!
	 *@brief Get a value.
	 *
	 *@param[in] key The key.
	 *
	 *@retval 0	@p key was not found.
	 *@retval !0	The pointer to the value string.
	 *
	 *@warning	The string is owned by the Parcel, and
	 *		must not be referenced after the Parcel has
	 *		been destroyed.
	 */
	const std::string*
	Get(const std::string& key) const;

	/*!
	 *@brief Extract an arbitrary type.
	 *
	 * Attempts to extract a T from a value string.
	 * Requires T to have a stream extractor operator.
	 *
	 *@param[in]  key	The key.
	 *@param[out] t		The object to extract.
	 *@param[in]  d		The default value if @p key is not found.
	 *
	 *@retval true	@p key was found and the type conversion was good.
	 *@retval false	@p key not found or the type conversion failed.
	 */
	template<typename T>
	bool
	Get(const std::string& key, T& t, const T& d = T()) const
	{
		const std::string*	v = Get(key);

		if (v == 0) {
			t = d;
			return(false);
		}

		std::istringstream s(*v);
		s >> t;
		if (!s) {
			t = d;
			return(false);
		}

		return(true);
	}

	/*!
	 *@brief Extract a string.
	 *
	 * Optimization to extract a std::string.
	 *
	 *@param[in]  key	The key.
	 *@param[out] s		The string to extract.
	 *@param[in]  d		The default value to if @p key is not found.
	 *
	 *@retval true	@p key was found.
	 *@retval false	@p key not found.
	 */
	bool
	Get(const std::string& key, std::string& s,
			const std::string& d = std::string()) const;

	/*!
	 *@brief Extract a nested Parcel.
	 *
	 * Extracts a nested Parcel, which was stored as a Package.
	 *
	 *@param[in]  key	The key.
	 *@param[out] p		The Parcel to extract.
	 *
	 *@retval true	@p key was found.
	 *@retval false	@p key not found.
	 */
	bool
	Get(const std::string& key, Parcel& p) const;

	/*!
	 *@brief Extract a nested Svec.
	 *
	 * Extracts a nested Svec, which was stored as a SvecPack.
	 *
	 *@param[in]  key	The key.
	 *@param[out] s		The Svec to extract.
	 *
	 *@retval true	@p key was found.
	 *@retval false	@p key not found.
	 */
	bool
	Get(const std::string& key, Svec& s) const;

	/*!
	 *@brief Extract a boolean value.
	 *
	 * Extracts a bool.  A @c true value is stored in a parcel
	 * by the presence of the key with any value.  A @c false
	 * value is indicated by the absence of the key.
	 *
	 *@param[in]  key	The key.
	 *@param[out] b		The bool to extract.
	 *
	 *@retval true	@p key was found.
	 *@retval false	@p key not found.
	 */
	bool
	Get(const std::string& key, bool& b) const
	{
		return(b = Present(key));
	}
	//@}

	//!@name Insert Key/Value Methods
	//@{
	/*!
	 *@brief Insert an arbitrary type value.
	 *
	 * Inserts a T into the Parcel.
	 * Requires T to have a stream inserter operator.
	 * Replaces the previous value for @p key, if any.
	 *
	 *@param[in] key	The key.
	 *@param[in] t		The value to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	template<typename T>
	Parcel&
	Set(const std::string& key, const T& t)
	{
		std::ostringstream s;

		s << t;
		Set(key, s.str());
		return(*this);
	}

	/*!
	 *@brief Insert a string value.
	 *
	 * Replaces the previous value for @P key, if any.
	 *
	 *@param[in] key	The key.
	 *@param[in] s		The string to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Set(const std::string& key, const std::string& s);

	/*!
	 *@brief Insert a C-string.
	 *
	 * Replaces the previous value for @p key, if any.
	 *
	 *@param[in] key	The key.
	 *@param[in] v		The string to insert.
	 *
	 *@return	A reference to this Parcel.
	 *
	 *@note		This method is required to prevent an implicit
	 *		conversion of char* to bool from erroneously
	 *		selecting Set(key, bool) instead!
	 */
	Parcel&
	Set(const std::string& key, const char* v)
	{
		return(Set(key, std::string(v)));
	}

	/*!
	 *@brief Insert a bool.
	 *
	 * Inserts a bool.  A @c true value is stored in a parcel
	 * by the presence of the key with an empty value.  A @c false
	 * value is indicated by the absence of the key.  This means
	 * inserting a @c false value will erase any previously stored
	 * value for @p key.
	 *
	 *@param[in] key	The key.
	 *@param[in] b		The bool to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&	
	Set(const std::string& key, bool b)
	{
		if (b)
			Set(key, std::string());
		else
			Erase(key);

		return(*this);
	}

	/*!
	 *@brief Insert a nested Parcel.
	 *
	 * Insert a nested Parcel, which is stored as a Package.
	 * Replaces the previous value for @p key, if any.
	 *
	 *@param[in] key	The key.
	 *@param[in] subp	The Parcel to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&	
	Set(const std::string& key, const Parcel& subp);

	/*!
	 *@brief Insert a nested Svec.
	 *
	 * Insert a nested Svec, which is stored as a SvecPack.
	 * Replaces the previous value for @p key, if any.
	 *
	 *@param[in] key	The key.
	 *@param[in] s		The Svec to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&	
	Set(const std::string& key, const Svec& s);
	//@}

	//!@name Merge Methods
	//@{
	/*!
	 *@brief Merge another Parcel.
	 *
	 * This method adds the contents of another Parcel to the
	 * current contents, possibly replacing current key/values if
	 * the other Parcel's keys match.
	 *
	 *@param[in] p	The other Parcel.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Merge(const Parcel& p);

	/*!
	 *@brief Operator form of merge a Parcel.
	 *
	 *@copydoc Merge(const Parcel& p)
	 */
	Parcel&
	operator+=(const Parcel& p)
	{
		return(Merge(p));
	}

	/*!
	 *@brief Merge a Package's contents.
	 *
	 * The Package is split into key/value pairs and they are added
	 * to the Parcel.  
	 *
	 *@param[in] p	The package to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Merge(const Package& p);

	/*!
	 *@brief Operator form of merge a Package.
	 *
	 *@copydoc Merge(const Package& p)
	 */
	Parcel&
	operator+=(const Package& p)
	{
		return(Merge(p));
	}

	/*!
	 *@brief Merge a Packet's contents.
	 *
	 * The Packet is collated into key/value pairs and is inserted
	 * into the Parcel.
	 *
	 *@param[in] p	The Packet to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Merge(const Packet& p);

	/*!
	 *@brief Operator form of merge a Packet.
	 *
	 *@copydoc Merge(const Packet& p)
	 */
	Parcel&
	operator+=(const Packet& p)
	{
		return(Merge(p));
	}

	/*!
	 *@brief Merge an IParmfile's contents.
	 *
	 *@param[in] ip	The IParmfile to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Merge(IParmfile& ip);

	/*!
	 *@brief Operator form of merge an IParmfile.
	 *
	 *@copydoc Merge(IParmfile& ip)
	 */
	Parcel&
	operator+=(IParmfile& ip)
	{
		return(Merge(ip));
	}
	//@}

	//!@name Assign Methods
	//@{
	/*!
	 *@brief Assign a Parcel.
	 *
	 * This method replaces the current contents of the Parcel
	 * with those of another Parcel.
	 *
	 *@param[in] p	The other Parcel.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Assign(const Parcel& p);

	/*!
	 *@brief Operator form of assign a Parcel.
	 *
	 *@copydoc Assign(const Parcel& p)
	 */
	Parcel&
	operator=(const Parcel& p)
	{
		return(Assign(p));
	}

	/*!
	 *@brief Assign a Package's contents into the Parcel.
	 *
	 * The Parcel is first cleared, then
	 * the Package inserted into the Parcel.
	 *
	 *@param[in] p	The Package to assign.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Assign(const Package& p)
	{
		Clear();
		return(Merge(p));
	}

	/*!
	 *@brief Operator form of assign a Package.
	 *
	 *@copydoc Assign(const Package& p)
	 */
	Parcel&
	operator=(const Package& p)
	{
		return(Assign(p));
	}

	/*!
	 *@brief Assign a Packet's contents into the Parcel.
	 *
	 * The Parcel is first cleared, then
	 * the Packet inserted into the Parcel.
	 *
	 *@param[in] p	The Packet to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Assign(const Packet& p)
	{
		Clear();
		return(Merge(p));
	}

	/*!
	 *@brief Operator form of assign a Packet.
	 *
	 *@copydoc Assign(const Packet& p)
	 */
	Parcel&
	operator=(const Packet& p)
	{
		return(Assign(p));
	}

	/*!
	 *@brief Assign an IParmfile's contents into the Parcel.
	 *
	 * The Parcel is first cleared, then
	 * the IParmfile inserted into the Parcel.
	 *
	 *@param[in] ip	The IParmfile to insert.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&
	Assign(IParmfile& ip)
	{
		Clear();
		return(Merge(ip));
	}

	/*!
	 *@brief Operator form of assign an IParmfile.
	 *
	 *@copydoc Assign(IParmfile& ip)
	 */
	Parcel&
	operator=(IParmfile& ip)
	{
		return(Assign(ip));
	}
	//@}

	//!@name Other Methods
	//@{
	/*!
	 *@brief Destructor
	 */
	~Parcel();

	/*!
	 *@brief Reset the Parcel to the empty state.
	 */
	Parcel&
	Clear(void);

	/*!
	 *@brief Remove a key/value, if present.
	 *
	 *@param[in] key	The key.
	 *
	 *@return	A reference to this Parcel.
	 */
	Parcel&	
	Erase(const std::string& key);

	/*!
	 *@brief Determine if a key is stored in the Parcel.
	 *
	 *@param[in] key	The key.
	 *
	 *@retval true	@p key was found.
	 *@retval false	@p key not found.
	 */
	bool
	Present(const std::string& key) const
	{
		const std::string*	p = Get(key);

		return(p != 0);
	}

	/*!
	 *@brief Access a key/value pair.
	 *
	 * Return an ItemRef object that represents a specific
	 * Parcel member.  Used for non-constant objects to provide
	 * read or write via subscripting.
	 *
	 *@param[in] key	The key.
	 *
	 *@return	An ItemRef for later use.
	 */
	ItemRef
	operator[](const std::string& key)
	{
		return(ItemRef(*this, key));
	}

	//
	// Constant objects with subscripts (reads) will use this version.
	//
	/*!
	 *@brief Access a constant key/value pair for reading.
	 *
	 *@param[in] key	The key.
	 *
	 *@retval 0	@p key was not found.
	 *@retval !0	The pointer to the value string.
	 */
	const std::string*
	operator[](const std::string& key) const
	{
		return(Get(key));
	}

	/*!
	 *@interface Visitor
	 *@brief Interface for enumerating a Parcel.
	 *
	 * A Visitor object's function call method @c operator() is called
	 * once for each key/value pair stored in the Parcel.
	 */
	class Visitor
	{
	public:
		/*!
		 *@brief Do something with a Parcel's key/value pair.
		 *
		 * Called once for each key/value pair while enumerating the
		 * Parcel.
		 *
		 *@param[in] key	The pair's key.
		 *@param[in] value	The pair's value.
		 *
		 *@retval true	Stop the enumeration.
		 *@retval false Continue the enumeration.
		 */
		virtual bool
		operator()(const std::string& key, const std::string& value)
				= 0;
	};

	/*!
	 *@brief Enumerate the contents of a Parcel.
	 *
	 *@param[in] v	The Visitor object.
	 *
	 *@retval true	The enumeration was prematurely terminated.
	 *@retval false The enumeration completed.
	 */
	bool
	Foreach(Visitor& v) const;

	/*!
	 *@brief Write a Parcel to an ostream with a selectable key/value
	 *	pair separator character.
	 *
	 *@param[in] os		The ostream.
	 *@param[in] sep	The separator character.
	 *
	 *@return	A reference to this Parcel.
	 */
	const Parcel&
	Output(std::ostream& os, char sep = ' ') const;
	//@}
};

/*!
 *@brief Send a Parcel to an ostream.
 *
 *@param[in] os	The ostream.
 *@param[in] p	The Parcel.
 *
 *@return	The ostream.
 */
inline std::ostream&
operator<<(std::ostream& os, const Parcel& p)
{
	p.Output(os);
	return(os);
}

}

#endif // LIBTOOLS_PARCEL_H
