/*!
 *@internal
 *@file		parcelp.h
 *@author	Craig Ruff
 *@date		July 31, 2002
 *@version	2
 *
 *@brief Parcel internal definitions.
 *
 * Copyright 2002 University Corporation for Atmospheric Research
 * All Rights Reserved
 */

#ifndef LIBTOOLS_PARCEL_P_H
#define LIBTOOLS_PARCEL_P_H

#include "parcel.h"
#include <map>

namespace Libtools {

/*!
 *@internal
 *@brief Internal Parcel representation.
 *
 * Parcels are implemented as reference counted, copy on modification,
 * string to string maps.
 */
class Parcel::Rep
{
private:
	typedef std::map<std::string, std::string> Pmap;
	typedef Pmap::iterator		PmapIter;
	typedef Pmap::const_iterator	PmapConstIter;

	Pmap		map;
	int		refCount;

	explicit
	Rep(const Pmap& m)
		: map(m), refCount(1)
	{
	}

	Rep(const Rep&);
	Rep& operator=(const Rep&);

	/*!
	 *@brief Get a copy, if necessary before modification.
	 *
	 *@return	The Rep to use.
	 */
	Rep*
	CopyOnWrite(void)
	{
		Rep	*t;

		if (refCount == 1) {
			t = this;

		} else {
			refCount--;
			t = new Rep(map);
		}

		return(t);
	}


public:
	/*!
	 *@brief Construct an empty representation.
	 */
	Rep()
		: map(), refCount(1)
	{}

	/*!
	 *brief Provide some of the standard forward iterators
	 *
	 *@note We do not provide non-constant iterators.
	 */
	typedef PmapConstIter	const_iterator;

	/*!
	 *@brief Forward the begin() iterator method.
	 */
	const_iterator
	begin(void) const
	{
		return(map.begin());
	}

	/*!
	 *@brief Forward the end() iterator method.
	 */
	const_iterator
	end(void) const
	{
		return(map.end());
	}

	/*!
	 *@brief Forward the find() method.
	 *
	 *@param[in] key	The key to erase.
	 *
	 *@return The iterator for the key/value pair or end() if not found.
	 */
	const_iterator
	find(const std::string& key) const
	{
		return(map.find(key));
	}

	/*!
	 *@brief Reset to the empty state.
	 *
	 *@return The representation to use.
	 */
	Rep*
	Clear(void)
	{
		Rep*	t = CopyOnWrite();

		t->map.clear();
		return(t);
	}

	/*!
	 *@brief Share the representation with another Parcel.
	 *
	 *@return The representation to use.
	 */
	Rep*
	Use(void)
	{
		refCount++;
		return(this);
	}

	/*!
	 *@brief Release the representation.
	 *
	 * A parent Parcel is going away.
	 *
	 *@note	This relies on the Parcel destructor
	 *	to delete us if Release returns zero!
	 *
	 *@return	The new reference count.
	 */
	int
	Release(void)
	{
		return(--refCount);
	}

	/*!
	 *@brief Remove a key/value from the map.
	 *
	 *@param[in] key	The key to erase.
	 *
	 *@return The representation to use.
	 */
	Rep*
	Erase(const std::string& key)
	{
		Rep*		t = CopyOnWrite();
		PmapIter	p = t->map.find(key);

		if (p != t->map.end())
			t->map.erase(p);

		return(t);
	}

	/*!
	 *@brief Place a key/value into the map.
	 *
	 *@param[in] key	The key to insert.
	 *@param[in] value	The corresponding value.
	 *
	 *@return The representation to use.
	 */
	Rep*
	Set(const std::string& key, const std::string& value)
	{
		Rep*	t = CopyOnWrite();

		t->map[key] = value;
		return(t);
	}
};

}

#endif // LIBTOOLS_PARCEL_P_H
